import asyncio
import functools
import httpx
import nest_asyncio
from typing import Callable, Any, TypeVar, Awaitable, Union, overload
try:
    from typing import ParamSpec
except ImportError:
    from typing_extensions import ParamSpec

T = TypeVar('T')
P = ParamSpec('P')

_sync_loop = None

@overload
def anywhere() -> Callable[[Callable[P, Awaitable[T]]], Callable[P, Union[Awaitable[T], T]]]: ...

def anywhere():
    """Decorator that allows an async function to run anywhere - sync or async context."""
    def decorator(async_func: Callable[P, Awaitable[T]]) -> Callable[P, Union[Awaitable[T], T]]:
        @functools.wraps(async_func)
        def wrapper(*args: P.args, **kwargs: P.kwargs) -> Union[Awaitable[T], T]:
            global _sync_loop

            try:
                nest_asyncio.apply()
            except ImportError:
                pass

            try:
                asyncio.get_running_loop()
                return async_func(*args, **kwargs) 
            except RuntimeError:
                if _sync_loop is None:
                    _sync_loop = asyncio.new_event_loop()
                asyncio.set_event_loop(_sync_loop) 
                try:
                    return _sync_loop.run_until_complete(async_func(*args, **kwargs))
                finally:
                    asyncio.set_event_loop(None)
        return wrapper
    return decorator

def ensure_client() -> Callable[[Callable[P, Awaitable[T]]], Callable[P, Awaitable[T]]]:
    """Decorator that ensures a valid httpx.AsyncClient exists before executing the method"""
    def decorator(async_func: Callable[P, Awaitable[T]]) -> Callable[P, Awaitable[T]]:
        @functools.wraps(async_func)
        async def wrapper(self: Any, *args: P.args, **kwargs: P.kwargs) -> T:
            if not self._client or self._client.is_closed:
                self._client = httpx.AsyncClient()
            return await async_func(self, *args, **kwargs)
        return wrapper
    return decorator

def require_task() -> Callable[[Callable[P, Awaitable[T]]], Callable[P, Awaitable[T]]]:
    """Decorator that ensures task has required attributes and valid client before execution"""
    def decorator(async_func: Callable[P, Awaitable[T]]) -> Callable[P, Awaitable[T]]:
        @functools.wraps(async_func)
        async def wrapper(self: Any, *args: P.args, **kwargs: P.kwargs) -> T:
            if not self.task_url:
                raise ValueError("Task URL not found")
            if not self._client:
                raise ValueError("Client not found")
            if not self._client._client or self._client._client.is_closed:
                self._client._client = httpx.AsyncClient()
            return await async_func(self, *args, **kwargs)
        return wrapper
    return decorator

def retry_on_429(max_retries: int = 3, initial_delay: float = 0.5) -> Callable[[Callable[P, Awaitable[T]]], Callable[P, Awaitable[T]]]:
    """Decorator that retries the request when encountering 429 Too Many Requests errors.
    
    Args:
        max_retries: Maximum number of retry attempts (default: 3)
        initial_delay: Initial delay in seconds, will be exponentially increased with jitter (default: 0.5)
    """
    def decorator(async_func: Callable[P, Awaitable[T]]) -> Callable[P, Awaitable[T]]:
        @functools.wraps(async_func)
        async def wrapper(*args: P.args, **kwargs: P.kwargs) -> T:
            import random 
            retries = 0
            while True:
                try:
                    return await async_func(*args, **kwargs)
                except httpx.HTTPStatusError as e:
                    if e.response.status_code != 429:
                        raise e
                    if retries >= max_retries:
                        print("Max retries reached")
                        raise e
                    retries += 1
                    delay = initial_delay * (2 ** retries)
                    # Use Retry-After header if available
                    retry_after = e.response.headers.get('Retry-After')
                    if retry_after:
                        try:
                            delay = float(retry_after)
                        except (ValueError, TypeError):
                            pass
                    jitter = random.uniform(0, 0.25) * delay
                    await asyncio.sleep(delay + jitter)
        return wrapper
    return decorator


from .configuration import Configuration
from .task_response import TaskResponse
from .auth import HeadersMixin
from abc import abstractmethod
from dotenv import load_dotenv
import httpx
import os
from pathlib import Path
from PIL import Image
from typing import BinaryIO, Union, Optional

class ChunkrBase(HeadersMixin):
    """Base class with shared functionality for Chunkr API clients.
    
    Args:
        url: The base URL of the Chunkr API. Defaults to the value of the CHUNKR_URL environment variable, or "https://api.chunkr.ai" if not set.
        api_key: The API key to use for authentication. Defaults to the value of the CHUNKR_API_KEY environment variable, or None if not set.
        raise_on_failure: Whether to raise an exception if the task fails. Defaults to False.
    """

    def __init__(self, url: str = None, api_key: str = None, raise_on_failure: bool = False):
        load_dotenv(override=True)
        self.url = url or os.getenv("CHUNKR_URL") or "https://api.chunkr.ai"
        self._api_key = api_key or os.getenv("CHUNKR_API_KEY")
        self.raise_on_failure = raise_on_failure
        
        if not self._api_key:
            raise ValueError(
                "API key must be provided either directly, in .env file, or as CHUNKR_API_KEY environment variable. You can get an api key at: https://www.chunkr.ai"
            )

        self.url = self.url.rstrip("/")
        self._client = httpx.AsyncClient()

    @abstractmethod
    def upload(
        self,
        file: Union[str, Path, BinaryIO, Image.Image],
        config: Configuration = None,
        filename: Optional[str] = None,
    ) -> TaskResponse:
        """Upload a file and wait for processing to complete.

        Args:
            file: The file to upload.
            config: Configuration options for processing. Optional.
            filename: The filename to use for the file. Optional.
            
        Examples:
        ```python
        # Upload from file path
        await chunkr.upload("document.pdf")

        # Upload from opened file
        with open("document.pdf", "rb") as f:
            await chunkr.upload(f)

        # Upload from URL
        await chunkr.upload("https://example.com/document.pdf")

        # Upload from base64 string (must include MIME type header)
        await chunkr.upload("data:application/pdf;base64,JVBERi0...", filename="document.pdf")

        # Upload an image
        from PIL import Image
        img = Image.open("photo.jpg")
        await chunkr.upload(img)
        ```
        Returns:
            TaskResponse: The completed task response
        """
        pass

    @abstractmethod
    def update(
        self, task_id: str, config: Configuration
    ) -> TaskResponse:
        """Update a task by its ID and wait for processing to complete.

        Args:
            task_id: The ID of the task to update
            config: Configuration options for processing. Optional.

        Returns:
            TaskResponse: The updated task response
        """
        pass

    @abstractmethod
    def create_task(
        self,
        file: Union[str, Path, BinaryIO, Image.Image],
        config: Configuration = None,
        filename: Optional[str] = None,
    ) -> TaskResponse:
        """Upload a file for processing and immediately return the task response. It will not wait for processing to complete. To wait for the full processing to complete, use `task.poll()`.

        Args:
            file: The file to upload.
            config: Configuration options for processing. Optional.
            filename: The filename to use for the file. Optional.
        Examples:
        ```
        # Upload from file path
        task = await chunkr.create_task("document.pdf")

        # Upload from opened file
        with open("document.pdf", "rb") as f:
            task = await chunkr.create_task(f)

        # Upload from URL
        task = await chunkr.create_task("https://example.com/document.pdf")

        # Upload from base64 string (must include MIME type header)
        task = await chunkr.create_task("data:application/pdf;base64,JVBERi0xLjcKCjEgMCBvYmo...", filename="document.pdf")

        # Upload an image
        from PIL import Image
        img = Image.open("photo.jpg")
        task = await chunkr.create_task(img)

        # Wait for the task to complete - this can be done when needed
        await task.poll()
        ```
        """
        pass

    @abstractmethod
    def update_task(
        self, task_id: str, config: Configuration
    ) -> TaskResponse:
        """Update a task by its ID and immediately return the task response. It will not wait for processing to complete. To wait for the full processing to complete, use `task.poll()`.

        Args:
            task_id: The ID of the task to update
            config: Configuration options for processing. Optional.

        Returns:
            TaskResponse: The updated task response
        """
        pass

    @abstractmethod
    def get_task(self, task_id: str, include_chunks: bool = True, base64_urls: bool = False) -> TaskResponse:
        """Get a task response by its ID.

        Args:
            task_id: The ID of the task to get
            include_chunks: Whether to include chunks in the output response. Defaults to True.
            base64_urls: Whether to return base64 encoded URLs. If false, the URLs will be returned as presigned URLs. Defaults to False.

        Returns:
            TaskResponse: The task response
        """
        pass

    @abstractmethod
    def delete_task(self, task_id: str) -> None:
        """Delete a task by its ID.

        Args:
            task_id: The ID of the task to delete
        """
        pass

    @abstractmethod
    def cancel_task(self, task_id: str) -> None:
        """Cancel a task by its ID.

        Args:
            task_id: The ID of the task to cancel
        """
        pass

    @abstractmethod
    def close(self) -> None:
        """Close the client connection. 
        This should be called when you're done using the client to properly clean up resources."""
        pass


from pydantic import BaseModel, Field, ConfigDict
from enum import Enum
from typing import Any, List, Optional, Union
from pydantic import field_validator

class GenerationStrategy(str, Enum):
    LLM = "LLM"
    AUTO = "Auto"

class CroppingStrategy(str, Enum):
    ALL = "All"
    AUTO = "Auto"

class EmbedSource(str, Enum):
    HTML = "HTML"
    MARKDOWN = "Markdown"
    LLM = "LLM"
    CONTENT = "Content"

class GenerationConfig(BaseModel):
    html: Optional[GenerationStrategy] = None
    llm: Optional[str] = None
    markdown: Optional[GenerationStrategy] = None
    crop_image: Optional[CroppingStrategy] = None
    embed_sources: Optional[List[EmbedSource]] = Field(default_factory=lambda: [EmbedSource.MARKDOWN])

class SegmentProcessing(BaseModel):
    model_config = ConfigDict(populate_by_name=True, alias_generator=str.title)

    caption: Optional[GenerationConfig] = Field(default=None, alias="Caption")
    footnote: Optional[GenerationConfig] = Field(default=None, alias="Footnote")
    formula: Optional[GenerationConfig] = Field(default=None, alias="Formula")
    list_item: Optional[GenerationConfig] = Field(default=None, alias="ListItem")
    page: Optional[GenerationConfig] = Field(default=None, alias="Page")
    page_footer: Optional[GenerationConfig] = Field(default=None, alias="PageFooter")
    page_header: Optional[GenerationConfig] = Field(default=None, alias="PageHeader")
    picture: Optional[GenerationConfig] = Field(default=None, alias="Picture")
    section_header: Optional[GenerationConfig] = Field(default=None, alias="SectionHeader")
    table: Optional[GenerationConfig] = Field(default=None, alias="Table")
    text: Optional[GenerationConfig] = Field(default=None, alias="Text")
    title: Optional[GenerationConfig] = Field(default=None, alias="Title")

class Tokenizer(str, Enum):
    WORD = "Word"
    CL100K_BASE = "Cl100kBase"
    XLM_ROBERTA_BASE = "XlmRobertaBase"
    BERT_BASE_UNCASED = "BertBaseUncased"

class TokenizerType(BaseModel):
    enum_value: Optional[Tokenizer] = None
    string_value: Optional[str] = None

    @classmethod
    def from_enum(cls, enum_value: Tokenizer) -> "TokenizerType":
        return cls(enum_value=enum_value)
    
    @classmethod
    def from_string(cls, string_value: str) -> "TokenizerType":
        return cls(string_value=string_value)
    
    def __str__(self) -> str:
        if self.enum_value is not None:
            return f"enum:{self.enum_value.value}"
        elif self.string_value is not None:
            return f"string:{self.string_value}"
        return ""
    
    model_config = ConfigDict(
        json_encoders={
            'TokenizerType': lambda v: v.model_dump()
        }
    )
    
    def model_dump(self, **kwargs):
        if self.enum_value is not None:
            return {"Enum": self.enum_value.value}
        elif self.string_value is not None:
            return {"String": self.string_value}
        return {}

class ChunkProcessing(BaseModel):
    ignore_headers_and_footers: Optional[bool] = True
    target_length: Optional[int] = None
    tokenizer: Optional[Union[TokenizerType, Tokenizer, str]] = None

    model_config = ConfigDict(
        arbitrary_types_allowed=True,
        json_encoders={
            TokenizerType: lambda v: v.model_dump()
        }
    )

    @field_validator('tokenizer', mode='before')
    def validate_tokenizer(cls, v):
        if v is None:
            return None
        
        if isinstance(v, TokenizerType):
            return v
        
        if isinstance(v, Tokenizer):
            return TokenizerType(enum_value=v)
        
        if isinstance(v, dict):
            if "Enum" in v:
                try:
                    return TokenizerType(enum_value=Tokenizer(v["Enum"]))
                except ValueError:
                    return TokenizerType(string_value=v["Enum"])
            elif "String" in v:
                return TokenizerType(string_value=v["String"])
            
        if isinstance(v, str):
            try:
                return TokenizerType(enum_value=Tokenizer(v))
            except ValueError:
                return TokenizerType(string_value=v)
                
        raise ValueError(f"Cannot convert {v} to TokenizerType")

class OcrStrategy(str, Enum):
    ALL = "All"
    AUTO = "Auto"

class SegmentationStrategy(str, Enum):
    LAYOUT_ANALYSIS = "LayoutAnalysis"
    PAGE = "Page"

class ErrorHandlingStrategy(str, Enum):
    FAIL = "Fail"
    CONTINUE = "Continue"

class BoundingBox(BaseModel):
    left: float
    top: float
    width: float
    height: float

class OCRResult(BaseModel):
    bbox: BoundingBox
    text: str
    confidence: Optional[float]

class SegmentType(str, Enum):
    CAPTION = "Caption"
    FOOTNOTE = "Footnote"
    FORMULA = "Formula"
    LIST_ITEM = "ListItem"
    PAGE = "Page"
    PAGE_FOOTER = "PageFooter"
    PAGE_HEADER = "PageHeader"
    PICTURE = "Picture"
    SECTION_HEADER = "SectionHeader"
    TABLE = "Table"
    TEXT = "Text"
    TITLE = "Title"

class Segment(BaseModel):
    bbox: BoundingBox
    content: str
    page_height: float
    llm: Optional[str] = None
    html: Optional[str] = None
    image: Optional[str] = None
    markdown: Optional[str] = None
    ocr: Optional[List[OCRResult]] = Field(default_factory=list)
    page_number: int
    page_width: float
    segment_id: str
    segment_type: SegmentType
    confidence: Optional[float]

class Chunk(BaseModel):
    chunk_id: str
    chunk_length: int
    segments: List[Segment]
    embed: Optional[str] = None

class OutputResponse(BaseModel):
    chunks: List[Chunk]
    file_name: Optional[str]
    page_count: Optional[int]
    pdf_url: Optional[str]

class Model(str, Enum):
    FAST = "Fast"
    HIGH_QUALITY = "HighQuality"

class Pipeline(str, Enum):
    AZURE = "Azure"
    CHUNKR = "Chunkr"

class Configuration(BaseModel):
    chunk_processing: Optional[ChunkProcessing] = None
    expires_in: Optional[int] = None
    error_handling: Optional[ErrorHandlingStrategy] = None
    high_resolution: Optional[bool] = None
    ocr_strategy: Optional[OcrStrategy] = None
    segment_processing: Optional[SegmentProcessing] = None
    segmentation_strategy: Optional[SegmentationStrategy] = None
    pipeline: Optional[Pipeline] = None
    
class OutputConfiguration(Configuration):
    input_file_url: Optional[str] = None
    # Deprecated
    json_schema: Optional[Any] = None
    model: Optional[Model] = None
    target_chunk_length: Optional[int] = None
    
class Status(str, Enum):
    STARTING = "Starting"
    PROCESSING = "Processing"
    SUCCEEDED = "Succeeded"
    FAILED = "Failed"
    CANCELLED = "Cancelled"



from .configuration import Configuration
import base64
import io
from pathlib import Path
from PIL import Image
from typing import Union, Tuple, BinaryIO, Optional

async def prepare_file(file: Union[str, Path, BinaryIO, Image.Image]) -> Tuple[Optional[str], str]:
    """Convert various file types into a tuple of (filename, file content).

    Args:
        file: Input file, can be:
            - URL string starting with http:// or https://
            - Base64 string
            - Local file path (will be converted to base64)
            - Opened binary file (will be converted to base64)
            - PIL/Pillow Image object (will be converted to base64)

    Returns:
        Tuple[Optional[str], str]: (filename, content) where content is either a URL or base64 string
        The filename may be None for URLs, base64 strings, and PIL Images

    Raises:
        FileNotFoundError: If the file path doesn't exist
        TypeError: If the file type is not supported
        ValueError: If the URL is invalid or unreachable
        ValueError: If the MIME type is unsupported
    """
    # Handle strings
    if isinstance(file, str):
        if file.startswith(('http://', 'https://')):
            return None, file
        try:
            base64.b64decode(file)
            return None, file
        except:
            try:
                file = Path(file)
            except:
                raise ValueError("File must be a valid path, URL, or base64 string")

    # Handle file paths - convert to base64
    if isinstance(file, Path):
        path = Path(file).resolve()
        if not path.exists():
            raise FileNotFoundError(f"File not found: {file}")
        
        with open(path, "rb") as f:
            file_content = f.read()
            file_ext = path.suffix.lower().lstrip('.')
            if not file_ext:
                raise ValueError("File must have an extension")
            base64_str = base64.b64encode(file_content).decode()
            return path.name, base64_str

    # Handle PIL Images - convert to base64
    if isinstance(file, Image.Image):
        img_byte_arr = io.BytesIO()
        format = file.format or "PNG"
        file.save(img_byte_arr, format=format)
        img_byte_arr.seek(0)
        base64_str = base64.b64encode(img_byte_arr.getvalue()).decode()
        return None, base64_str

    # Handle file-like objects - convert to base64
    if hasattr(file, "read") and hasattr(file, "seek"):
        file.seek(0)
        file_content = file.read()
        name = getattr(file, "name", "document")
        file_ext = Path(name).suffix.lower().lstrip('.')
        if not file_ext:
            raise ValueError("File must have an extension")
        base64_str = base64.b64encode(file_content).decode()
        return Path(name).name, base64_str

    raise TypeError(f"Unsupported file type: {type(file)}")


async def prepare_upload_data(
    file: Optional[Union[str, Path, BinaryIO, Image.Image]] = None,
    filename: Optional[str] = None,
    config: Optional[Configuration] = None,
) -> dict:
    """Prepare data dictionary for upload.

    Args:
        file: The file to upload
        config: Optional configuration settings
        client: HTTP client for downloading remote files

    Returns:
        dict: JSON-serializable data dictionary ready for upload
    """
    data = {}
    if file:
        processed_filename, processed_file = await prepare_file(file)
        data["file"] = processed_file
        data["file_name"] = filename or processed_filename

    if config:
        data.update(config.model_dump(mode="json", exclude_none=True))
        
    return data



import pytest
from pathlib import Path
from PIL import Image
import asyncio
import base64

from chunkr_ai import Chunkr
from chunkr_ai.models import (
    Configuration,
    GenerationStrategy,
    GenerationConfig,
    OcrStrategy,
    Pipeline,
    SegmentationStrategy,
    SegmentProcessing,
    ChunkProcessing,
    TaskResponse,
    EmbedSource,
    ErrorHandlingStrategy,
    Tokenizer,
)

@pytest.fixture
def sample_path():
    return Path("tests/files/test.pdf")

@pytest.fixture
def sample_image():
    return Image.open("tests/files/test.jpg")

@pytest.fixture
def sample_url():
    return "https://chunkr-web.s3.us-east-1.amazonaws.com/landing_page/input/science.pdf"

@pytest.fixture
def client():
    client = Chunkr()
    yield client

@pytest.fixture
def markdown_embed_config():
    return Configuration(
        segment_processing=SegmentProcessing(
            page=GenerationConfig(
                html=GenerationStrategy.LLM,
                markdown=GenerationStrategy.LLM,
                embed_sources=[EmbedSource.MARKDOWN]
            )
        ),
    )

@pytest.fixture
def html_embed_config():
    return Configuration(
        segment_processing=SegmentProcessing(
            page=GenerationConfig(
                html=GenerationStrategy.LLM,
                markdown=GenerationStrategy.LLM,
                embed_sources=[EmbedSource.HTML]
            )
        ),
    )

@pytest.fixture
def multiple_embed_config():
    return Configuration(
        segment_processing=SegmentProcessing(
            page=GenerationConfig(
                html=GenerationStrategy.LLM,
                markdown=GenerationStrategy.LLM,
                llm="Generate a summary of this content",
                embed_sources=[EmbedSource.MARKDOWN, EmbedSource.LLM, EmbedSource.HTML]
            )
        ),
    )
    
@pytest.fixture
def word_tokenizer_string_config():
    return Configuration(
        chunk_processing=ChunkProcessing(
            tokenizer="Word"
        ),
    )

@pytest.fixture
def word_tokenizer_config():
    return Configuration(
        chunk_processing=ChunkProcessing(
            tokenizer=Tokenizer.WORD
        ),
    )

@pytest.fixture
def cl100k_tokenizer_config():
    return Configuration(
        chunk_processing=ChunkProcessing(
            tokenizer=Tokenizer.CL100K_BASE
        ),
    )

@pytest.fixture
def custom_tokenizer_config():
    return Configuration(
        chunk_processing=ChunkProcessing(
            tokenizer="Qwen/Qwen-tokenizer"
        ),
    )

@pytest.fixture
def xlm_roberta_with_html_content_config():
    return Configuration(
        chunk_processing=ChunkProcessing(
            tokenizer=Tokenizer.XLM_ROBERTA_BASE
        ),
        segment_processing=SegmentProcessing(
            page=GenerationConfig(
                html=GenerationStrategy.LLM,
                markdown=GenerationStrategy.LLM,
                embed_sources=[EmbedSource.HTML, EmbedSource.CONTENT]
            )
        ),
    )

@pytest.mark.asyncio
async def test_send_file_path(client, sample_path):
    response = await client.upload(sample_path)
    assert response.task_id is not None
    assert response.status == "Succeeded"
    assert response.output is not None

@pytest.mark.asyncio
async def test_send_file_url(client, sample_url):
    response = await client.upload(sample_url)
    assert response.task_id is not None
    assert response.status == "Succeeded"
    assert response.output is not None

@pytest.mark.asyncio
async def test_send_file_path_str(client, sample_path):
    response = await client.upload(str(sample_path))
    assert response.task_id is not None
    assert response.status == "Succeeded"
    assert response.output is not None

@pytest.mark.asyncio
async def test_send_opened_file(client, sample_path):
    with open(sample_path, "rb") as f:
        response = await client.upload(f)
    assert response.task_id is not None
    assert response.status == "Succeeded"
    assert response.output is not None

@pytest.mark.asyncio
async def test_send_pil_image(client, sample_image):
    response = await client.upload(sample_image)
    assert response.task_id is not None
    assert response.status == "Succeeded"
    assert response.output is not None
    assert response.output is not None

@pytest.mark.asyncio
async def test_ocr_auto(client, sample_path):
    response = await client.upload(sample_path, Configuration(ocr_strategy=OcrStrategy.AUTO))
    assert response.task_id is not None
    assert response.status == "Succeeded"
    assert response.output is not None

@pytest.mark.asyncio
async def test_expires_in(client, sample_path):
    response = await client.upload(sample_path, Configuration(expires_in=10))
    assert response.task_id is not None
    assert response.status == "Succeeded"
    assert response.output is not None
    await asyncio.sleep(11)
    try:
        await client.get_task(response.task_id)
        assert False
    except Exception as e:
        print(e)
        assert True

@pytest.mark.asyncio
async def test_chunk_processing(client, sample_path):
    response = await client.upload(
        sample_path,
        Configuration(chunk_processing=ChunkProcessing(target_length=1024)),
    )
    assert response.task_id is not None
    assert response.status == "Succeeded"
    assert response.output is not None

@pytest.mark.asyncio
async def test_segmentation_strategy_page(client, sample_path):
    response = await client.upload(
        sample_path, Configuration(segmentation_strategy=SegmentationStrategy.PAGE)
    )
    assert response.task_id is not None
    assert response.status == "Succeeded"
    assert response.output is not None

@pytest.mark.asyncio
async def test_page_llm_html(client, sample_path):
    response = await client.upload(
        sample_path,
        Configuration(
            segmentation_strategy=SegmentationStrategy.PAGE,
            segment_processing=SegmentProcessing(
                page=GenerationConfig(html=GenerationStrategy.LLM)
            ),
        ),
    )
    assert response.task_id is not None
    assert response.status == "Succeeded"
    assert response.output is not None

@pytest.mark.asyncio
async def test_page_llm(client, sample_path):
    configuration = Configuration(
        segmentation_strategy=SegmentationStrategy.PAGE,
        segment_processing=SegmentProcessing(
            page=GenerationConfig(
                html=GenerationStrategy.LLM, markdown=GenerationStrategy.LLM
            )
        ),
    )
    response = await client.upload(sample_path, configuration)
    assert response.task_id is not None
    assert response.status == "Succeeded"
    assert response.output is not None

@pytest.mark.asyncio
async def test_delete_task(client, sample_path):
    response = await client.upload(sample_path)
    await client.delete_task(response.task_id)
    with pytest.raises(Exception):
        await client.get_task(response.task_id)

@pytest.mark.asyncio
async def test_delete_task_direct(client, sample_path):
    task = await client.upload(sample_path)
    assert isinstance(task, TaskResponse)
    assert task.task_id is not None
    assert task.status == "Succeeded"
    assert task.output is not None
    await client.delete_task(task.task_id)
    with pytest.raises(Exception):
        await client.get_task(task.task_id)

@pytest.mark.asyncio
async def test_cancel_task(client, sample_path):
    response = await client.create_task(sample_path)
    assert response.status == "Starting"
    await client.cancel_task(response.task_id)
    assert (await client.get_task(response.task_id)).status == "Cancelled"

@pytest.mark.asyncio
async def test_cancel_task_direct(client, sample_path):
    task = await client.create_task(sample_path)
    assert isinstance(task, TaskResponse)
    assert task.status == "Starting"
    await task.cancel()
    assert task.status == "Cancelled"

@pytest.mark.asyncio
async def test_update_task(client, sample_path):
    original_config = Configuration(
        segmentation_strategy=SegmentationStrategy.LAYOUT_ANALYSIS,
    )
    new_config = Configuration(
        segmentation_strategy=SegmentationStrategy.PAGE,
    )
    task = await client.upload(sample_path, original_config)
    task = await client.update(task.task_id, new_config)
    assert task.status == "Succeeded"
    assert task.output is not None
    assert task.configuration.segmentation_strategy == SegmentationStrategy.PAGE

@pytest.mark.asyncio
async def test_update_task_direct(client, sample_path):
    original_config = Configuration(
        segmentation_strategy=SegmentationStrategy.LAYOUT_ANALYSIS,
    )
    new_config = Configuration(
        segmentation_strategy=SegmentationStrategy.PAGE,
    )
    task = await client.upload(sample_path, original_config)
    task = await task.update(new_config)
    assert task.status == "Succeeded"
    assert task.output is not None
    assert task.configuration.segmentation_strategy == SegmentationStrategy.PAGE

@pytest.mark.asyncio
async def test_pipeline_type_azure(client, sample_path):
    response = await client.upload(sample_path, Configuration(pipeline=Pipeline.AZURE))
    assert response.task_id is not None
    assert response.status == "Succeeded"
    assert response.output is not None
    
@pytest.mark.asyncio
async def test_pipeline_type_azure(client, sample_path):
    response = await client.upload(sample_path, Configuration(pipeline=Pipeline.CHUNKR))
    assert response.task_id is not None
    assert response.status == "Succeeded"
    assert response.output is not None

@pytest.mark.asyncio
async def test_client_lifecycle(client, sample_path):
    response1 = await client.upload(sample_path)
    await client.close()
    response2 = await client.upload(sample_path)
    assert response1.task_id is not None
    assert response2.task_id is not None

@pytest.mark.asyncio
async def test_task_operations_after_client_close(client, sample_path):
    task = await client.create_task(sample_path)
    await client.close()
    result = await task.poll()
    assert result.status == "Succeeded"

@pytest.mark.asyncio
async def test_send_base64_file(client, sample_path):
    # Read file and convert to base64
    with open(sample_path, "rb") as f:
        base64_content = base64.b64encode(f.read()).decode('utf-8')
    response = await client.upload(base64_content)
    assert response.task_id is not None
    assert response.status == "Succeeded"
    assert response.output is not None

@pytest.mark.asyncio
async def test_send_base64_file_with_data_url(client, sample_path):
    with open(sample_path, "rb") as f:
        base64_content = base64.b64encode(f.read()).decode('utf-8')
    response = await client.upload(f"data:application/pdf;base64,{base64_content}")
    assert response.task_id is not None
    assert response.status == "Succeeded"
    assert response.output is not None

@pytest.mark.asyncio
async def test_send_base64_file_with_filename(client, sample_path):
    # Read file and convert to base64
    with open(sample_path, "rb") as f:
        base64_content = base64.b64encode(f.read()).decode('utf-8')
    
    response = await client.upload(base64_content, filename="test.pdf")
    assert response.task_id is not None
    assert response.status == "Succeeded"
    assert response.output is not None
    
@pytest.mark.asyncio
async def test_output_files_no_dir(client, sample_path, tmp_path):
    task = await client.upload(sample_path)
    
    html_file = tmp_path / "output.html"
    md_file = tmp_path / "output.md"
    content_file = tmp_path / "output.txt"
    json_file = tmp_path / "output.json"
    
    task.html(html_file)
    task.markdown(md_file)
    task.content(content_file)
    task.json(json_file)
    
    assert html_file.exists()
    assert md_file.exists()
    assert content_file.exists()
    assert json_file.exists()

@pytest.mark.asyncio
async def test_output_files_with_dirs(client, sample_path, tmp_path):
    task = await client.upload(sample_path)
    
    nested_dir = tmp_path / "nested" / "output" / "dir"
    html_file = nested_dir / "output.html"
    md_file = nested_dir / "output.md"
    content_file = nested_dir / "output.txt"
    json_file = nested_dir / "output.json"
    
    task.html(html_file)
    task.markdown(md_file)
    task.content(content_file)
    task.json(json_file)

    assert html_file.exists()
    assert md_file.exists()
    assert content_file.exists()
    assert json_file.exists()

@pytest.mark.asyncio
async def test_embed_sources_markdown_only(client, sample_path, markdown_embed_config):
    response = await client.upload(sample_path, markdown_embed_config)
    assert response.task_id is not None
    assert response.status == "Succeeded"
    assert response.output is not None
    # Check the first chunk to verify embed exists
    if response.output.chunks:
        chunk = response.output.chunks[0]
        assert chunk.embed is not None

@pytest.mark.asyncio
async def test_embed_sources_html_only(client, sample_path, html_embed_config):
    response = await client.upload(sample_path, html_embed_config)
    assert response.task_id is not None
    assert response.status == "Succeeded"
    assert response.output is not None

@pytest.mark.asyncio
async def test_embed_sources_multiple(client, sample_path, multiple_embed_config):
    response = await client.upload(sample_path, multiple_embed_config)
    assert response.task_id is not None
    assert response.status == "Succeeded"
    assert response.output is not None

@pytest.mark.asyncio
async def test_tokenizer_word(client, sample_path, word_tokenizer_config):
    response = await client.upload(sample_path, word_tokenizer_config)
    assert response.task_id is not None
    assert response.status == "Succeeded"
    assert response.output is not None
    if response.output.chunks:
        for chunk in response.output.chunks:
            # Word tokenizer should result in chunks with length close to target
            assert chunk.chunk_length > 0
            assert chunk.chunk_length <= 600  # Allow some flexibility

@pytest.mark.asyncio
async def test_tokenizer_cl100k(client, sample_path, cl100k_tokenizer_config):
    response = await client.upload(sample_path, cl100k_tokenizer_config)
    assert response.task_id is not None
    assert response.status == "Succeeded"
    assert response.output is not None

@pytest.mark.asyncio
async def test_tokenizer_custom_string(client, sample_path, custom_tokenizer_config):
    response = await client.upload(sample_path, custom_tokenizer_config)
    assert response.task_id is not None
    assert response.status == "Succeeded"
    assert response.output is not None

@pytest.mark.asyncio
async def test_embed_sources_with_different_tokenizer(client, sample_path, xlm_roberta_with_html_content_config):
    response = await client.upload(sample_path, xlm_roberta_with_html_content_config)
    assert response.task_id is not None
    assert response.status == "Succeeded"
    assert response.output is not None

@pytest.mark.asyncio
async def test_error_handling_continue(client, sample_path):
    response = await client.upload(sample_path, Configuration(error_handling=ErrorHandlingStrategy.CONTINUE))
    assert response.task_id is not None
    assert response.status == "Succeeded"
    assert response.output is not None



from .api.configuration import (
    BoundingBox,
    Chunk,
    ChunkProcessing,
    Configuration,
    CroppingStrategy,
    EmbedSource,
    ErrorHandlingStrategy,
    GenerationStrategy,
    GenerationConfig,
    Model,
    OCRResult,
    OcrStrategy,
    OutputResponse,
    Segment,
    SegmentProcessing,
    SegmentType,
    SegmentationStrategy,
    Status,
    Pipeline,
    Tokenizer,
    TokenizerType,
)
from .api.task_response import TaskResponse

__all__ = [
    "BoundingBox",
    "Chunk",
    "ChunkProcessing",
    "Configuration",
    "CroppingStrategy",
    "EmbedSource",
    "ErrorHandlingStrategy",
    "GenerationConfig",
    "GenerationStrategy",
    "Model",
    "OCRResult",
    "OcrStrategy",
    "OutputResponse",
    "Segment",
    "SegmentProcessing",
    "SegmentType",
    "SegmentationStrategy",
    "Status",
    "TaskResponse",
    "Pipeline",
    "Tokenizer",
    "TokenizerType",
]



from typing import Optional, runtime_checkable, Protocol
from httpx import AsyncClient


@runtime_checkable
class ChunkrClientProtocol(Protocol):
    """Protocol defining the interface for Chunkr clients"""
    
    raise_on_failure: bool = True
    _client: Optional[AsyncClient] = None

    def _headers(self) -> dict:
        """Return headers required for API requests"""
        ...



from datetime import datetime
from typing import TypeVar, Optional, Generic
from pydantic import BaseModel, PrivateAttr
import asyncio
import json
import os
import httpx

from .configuration import Configuration, OutputConfiguration, OutputResponse, Status
from .protocol import ChunkrClientProtocol
from .misc import prepare_upload_data
from .decorators import anywhere, require_task, retry_on_429

T = TypeVar("T", bound="TaskResponse")

class TaskResponse(BaseModel, Generic[T]):
    configuration: OutputConfiguration
    created_at: datetime
    expires_at: Optional[datetime] = None
    finished_at: Optional[datetime] = None
    message: Optional[str] = None
    output: Optional[OutputResponse] = None
    started_at: Optional[datetime] = None
    status: Status
    task_id: str
    task_url: Optional[str] = None
    include_chunks: bool = False
    _base64_urls: bool = False
    _client: Optional[ChunkrClientProtocol] = PrivateAttr(default=None)

    def with_client(self, client: ChunkrClientProtocol, include_chunks: bool = False, base64_urls: bool = False) -> T:
        self._client = client
        self.include_chunks = include_chunks
        self._base64_urls = base64_urls
        return self

    def _check_status(self) -> Optional[T]:
        """Helper method to check task status and handle completion/failure"""
        if self.status == "Failed":
            if getattr(self._client, 'raise_on_failure', True):
                raise ValueError(self.message)
            return self
        if self.status not in ("Starting", "Processing"):
            return self
        return None

    @require_task()
    async def _poll_request(self) -> dict:
        try:
            r = await self._client._client.get(
                self.task_url, headers=self._client._headers()
            )
            r.raise_for_status()
            return r.json()
        except (ConnectionError, TimeoutError, OSError, 
                httpx.ReadTimeout, httpx.ConnectTimeout, 
                httpx.WriteTimeout, httpx.PoolTimeout,
                httpx.ConnectError, httpx.ReadError, 
                httpx.NetworkError) as e:
            print(f"Connection error while polling the task: {str(e)}\nretrying...")
            await asyncio.sleep(0.5)
            return await self._poll_request() 
        except Exception as e:
            raise e

    @anywhere()
    async def poll(self) -> T:
        """Poll the task for completion."""
        while True:
            j = await self._poll_request()
            updated = TaskResponse(**j).with_client(self._client)
            self.__dict__.update(updated.__dict__)
            if res := self._check_status():
                return res
            await asyncio.sleep(0.5)

    @anywhere()
    @require_task()
    @retry_on_429()
    async def update(self, config: Configuration) -> T:
        """Update the task configuration."""
        data = await prepare_upload_data(None, None, config)
        r = await self._client._client.patch(
            f"{self.task_url}/parse",
            json=data,
            headers=self._client._headers()
        )
        r.raise_for_status()
        updated = TaskResponse(**r.json()).with_client(self._client)
        self.__dict__.update(updated.__dict__)
        return await self.poll()

    @anywhere()
    @require_task()
    async def delete(self) -> T:
        """Delete the task."""
        r = await self._client._client.delete(
            self.task_url, headers=self._client._headers()
        )
        r.raise_for_status()
        return self

    @anywhere()
    @require_task()
    async def cancel(self) -> T:
        """Cancel the task."""
        r = await self._client._client.get(
            f"{self.task_url}/cancel", headers=self._client._headers()
        )
        r.raise_for_status()
        return await self.poll()

    def _write_to_file(self, content: str | dict, output_file: str, is_json: bool = False) -> None:
        """Helper method to write content to a file
        
        Args:
            content: Content to write (string or dict for JSON)
            output_file: Path to save the content
            is_json: Whether the content should be written as JSON
        """
        class DateTimeEncoder(json.JSONEncoder):
            def default(self, obj):
                if isinstance(obj, datetime):
                    return obj.isoformat()
                return super().default(obj)
        if output_file:
            directory = os.path.dirname(output_file)
            if directory:
                os.makedirs(directory, exist_ok=True)
            with open(output_file, "w", encoding="utf-8") as f:
                if is_json:
                    json.dump(content, f, cls=DateTimeEncoder, indent=2)
                else:
                    f.write(content)

    def html(self, output_file: str = None) -> str:
        """Get the full HTML of the task
        
        Args:
            output_file (str, optional): Path to save the HTML content. Defaults to None.
        """
        content = self._get_content("html")
        self._write_to_file(content, output_file)
        return content

    def markdown(self, output_file: str = None) -> str:
        """Get the full markdown of the task
        
        Args:
            output_file (str, optional): Path to save the markdown content. Defaults to None.
        """
        content = self._get_content("markdown", separator="\n\n")
        self._write_to_file(content, output_file)
        return content

    def content(self, output_file: str = None) -> str:
        """Get the full content of the task
        
        Args:
            output_file (str, optional): Path to save the content. Defaults to None.
        """
        content = self._get_content("content")
        self._write_to_file(content, output_file)
        return content
    
    def json(self, output_file: str = None) -> dict:
        """Get the full task data as JSON
        
        Args:
            output_file (str, optional): Path to save the task data as JSON. Defaults to None.
        """
        data = self.model_dump()
        self._write_to_file(data, output_file, is_json=True)
        return data

    def _get_content(self, t: str, separator: str = "\n") -> str:
        if not self.output:
            return ""
        parts = []
        for c in self.output.chunks:
            for s in c.segments:
                v = getattr(s, t)
                if v:
                    parts.append(v)
        return separator.join(parts)



class HeadersMixin:
    """Mixin class for handling authorization headers"""

    def get_api_key(self) -> str:
        """Get the API key"""
        if not hasattr(self, "_api_key") or not self._api_key:
            raise ValueError("API key not set")
        return self._api_key

    def _headers(self) -> dict:
        """Generate authorization headers"""
        return {"Authorization": self.get_api_key()}
